library("gdata")
library("prodlim") # row.match
setwd("/home/abn/Escritorio/TFG-Estadística/Ejecuciones")
########################################################################
###########################  FUNCIONES #################################
########################################################################

Dibujar_grafo <- function(cx,cy,arcos){
  plot(cx,cy,type="n")
  text(cx,cy, labels = seq(1,length(cx),1),cex=.8)
  edges <- as.vector(t(which(arcos==1,arr.ind=TRUE)))
  for(i in seq(1,length(edges),2)){
    lines(c(cx[edges[i]],cx[edges[i+1]]),c(cy[edges[i]],cy[edges[i+1]]),type = "l",col="blue")
  }
}


Es_factible <- function(arcos,H,raiz){
  # Primero en profundidad desde el nodo raíz manteniendo un contador de saltos
  n_nodos <- length(arcos[1,])
  nodos_visitados <- rep(0,n_nodos)
  arcos_copia <- arcos
  # stack se tratará como una pila ( guarda los nodos a explorar)
  # hops se tratará como una pila ( guarda los saltos de los nodos a explorar)
  stack <- c(raiz)
  hops <- c(0)
  nodo_anterior <- 0
  while(length(stack)!=0){
    nodo_actual <- stack[length(stack)]
    #print(paste("Nodo actual: ",nodo_actual,sep=""))
    saltos_actuales <- hops[length(hops)]
    stack <- stack[-length(stack)]
    hops <- hops[-length(hops)]
    if(nodos_visitados[nodo_actual] == 0){
      nodos_visitados[nodo_actual] = 1
      for(i in which(arcos_copia[nodo_actual,]==1) ){
        stack<-c(stack,i)
        hops <-c(hops,saltos_actuales+1)
        # eliminamos el arco que acabamos de usar
        arcos_copia[nodo_actual,i] = 0
        arcos_copia[i,nodo_actual] = 0
        #########################################
        if((saltos_actuales+1)>H){
          # Más saltos que los permitidos
          #print("Más saltos que los permitidos")
          return(FALSE)
        }
      }
      # print("Stack ")
      # print(stack)
      # print("Hops ")
      # print(hops)
    }
    else{
      # Hay un ciclo
      #print("Hay un ciclo")
      return(FALSE)
    }
  }
  if(sum(nodos_visitados)<n_nodos){
    # Existen subciclos
    #print("Hay subciclos")
    return(FALSE)
  }
  #print("Todo correcto")
  return(TRUE)
}


get_CaminoySubtree <- function(arcos,raiz,nodo_obj){
  # Primero en profundidad desde el nodo objetivo
  n_nodos <- length(arcos[1,])
  nodos_visitados <- rep(0,n_nodos)
  nodos_visitados[raiz] <- 1
  arcos_copia <- arcos
  flag_break <- 0
  # stack se tratará como una pila ( guarda los nodos a explorar)
  solucion <- c()
  hijos <- c()
  stack <- c(nodo_obj)
  while(length(stack)!=0){
    nodo_actual <- stack[length(stack)]
    stack <- stack[-length(stack)]
    if(nodos_visitados[nodo_actual] == 0){
      solucion<-c(solucion,nodo_actual)
      nodos_visitados[nodo_actual] = 1
      n_hijos = 0
      flag_up = 0
      for(i in which(arcos[nodo_actual,]==1)){
        stack<-c(stack,i)
        n_hijos = n_hijos+1
        flag_up = 1
      }
    }
  }
  solucion<-c(solucion,raiz)
  return(solucion)
}

REW <- function(arcos, tamRCL){
  final = 0
  n_puntos <- length(arcos[1,])
  savings <-array(NA,dim=c(n_puntos,n_puntos,n_puntos)) # NA significa que no es posible el cambio
  while(final==0){
    # Se puede mejorar el tiempo que tarda si no se recalculan todos los savings en cada paso
    
    #edges <- as.vector(t(which(arcos==1,arr.ind=TRUE)))
    for(i in  1:n_puntos){
      # Arco actual
      #punto1_1=edges[i]
      punto2_1=i
      # Cambio para reducir los tiempos
      ###############################################
      # Solo los que esten conectados a la raiz
      if(punto2_1 == raiz){next}
      if(arcos[punto2_1,raiz] != 1){next}
      # Buscamos el arco por el que cambiarlo
      
      
      
      ###############################################
      # Buscamos el arco por el que cambiarlo
      non_edges <- which(arcos==0,arr.ind=TRUE)
      # Quitamos los arcos de un nodo a si mismo
      non_edges <- non_edges[non_edges[,1]!=non_edges[,2],]
      subtree <- get_CaminoySubtree(arcos,raiz,punto2_1)
      #print(paste("Se evalua ",punto2_1,sep=""))
      non_edges <- non_edges[non_edges[,2] %in% subtree,] # <- que pertenezca al subárbol
      non_edges <- matrix(non_edges,nrow=length(non_edges)/2)
      if(length(non_edges) == 0){ next}
      non_edges <- non_edges[!(non_edges[,2] %in% raiz),]# <- pero no es la raiz
      non_edges <- matrix(non_edges,nrow=length(non_edges)/2)
      if(length(non_edges) == 0){ next}
      non_edges <- non_edges[!(non_edges[,1] %in% raiz),]# <- pero no es la raiz
      if(length(non_edges) == 0){ next}
      non_edges <- matrix(non_edges,nrow=length(non_edges)/2)
      if(length(non_edges)>0){
        for(j in  1:length(non_edges[,1])){
          punto1_2=non_edges[j,1]
          punto2_2=non_edges[j,2]
          if(is.na(savings[punto2_1,punto1_2,punto2_2])){
            #Probar que el cambio da una solución factible
            arcos[raiz,punto2_1] = 0
            arcos[punto2_1,raiz] = 0
            arcos[punto1_2,punto2_2] = 1
            arcos[punto2_2,punto1_2] = 1
            
            # Si es factible guardar ahorro
            if(Es_factible(arcos,H,raiz)){
              #print(paste("Arco (",punto1_1,",",punto2_1,") por (",punto1_2,",",punto2_1,")",sep = ""))
              savings[punto2_1,punto1_2,punto2_2] <- costes[raiz,punto2_1] - costes[punto1_2,punto2_2]
              #print(paste("Ahorro = ",savings[punto1_1,punto2_1,punto1_2], sep=""))
            }
            
            # Revertir el cambio
            arcos[raiz,punto2_1] = 1
            arcos[punto2_1,raiz] = 1
            arcos[punto1_2,punto2_2] = 0
            arcos[punto2_2,punto1_2] = 0
          }
        }
      }
    }
    # Tenemos en savings los ahorros calculados
    # Encontrar los tamRCL-ésimos mejores ahorros
    max_savings = c()
    cambio_i <- c()
    cambio_j <- c()
    cambio_k <- c()
    for(i in 1:n_puntos){
      for(j in 1:n_puntos){
        for(k in 1:n_puntos){
          if(length(max_savings)<tamRCL){
            if(!is.na(savings[i,j,k]) && savings[i,j,k]>0){
              max_savings <- c(max_savings,savings[i,j,k])
              cambio_i <- c(cambio_i,i)
              cambio_j <- c(cambio_j,j)
              cambio_k <- c(cambio_k,k)
            }
          }
          if(length(max_savings)==tamRCL && isTRUE(savings[i,j,k]>min(max_savings))){
            posicion = which(max_savings == min(max_savings))
            max_savings[posicion] <- savings[i,j,k]
            cambio_i[posicion] <- i
            cambio_j[posicion] <- j
            cambio_k[posicion] <- k
          }
        }
      }
    }
    
    flagbreak = 1
    for(i in 1:tamRCL){
      if(isTRUE(max_savings[i] > 0) ){
        # Hay mejora
        flagbreak = 0
      }
    }
    if(flagbreak == 1) {break;}
    
    
    #print(paste("Arco (",punto1_1,",",punto2_1,") por (",punto1_2,",",punto2_1,")",sep = ""))
    
    #print("SAVINGS")
    #print(savings[25,3,1])
    #print("MAX-SAVINGS")
    #print(max_savings)
    
    eleccion = length(max_savings)+1
    while(eleccion == length(max_savings)+1){
      eleccion = trunc(runif(1,1,length(max_savings)+1))
    }
    #print(paste("Arco (",punto1_1,",",punto2_1,") por (",punto1_2,",",punto2_1,")",sep = ""))
    
    arcos[raiz,cambio_i[eleccion]] = 0
    arcos[cambio_i[eleccion],raiz] = 0
    arcos[cambio_j[eleccion],cambio_k[eleccion]] = 1
    arcos[cambio_k[eleccion],cambio_j[eleccion]] = 1
    
    #Dibujar_grafo(cx,cy,arcos)
    
    #Se marcan los savings que desaparecen o han de ser actualizados
    
    actualizantes <- get_CaminoySubtree(arcos,raiz,cambio_j[eleccion])
    savings[,actualizantes,] = NA
    savings[,,actualizantes] = NA
    savings[actualizantes,,] = NA
    
  }
  return(arcos)
}


get_Grafo <- function(niveles,costes){
  n_puntos <- length(niveles)
  raiz <- which(niveles == 0)
  resultado <- matrix (0,n_puntos,n_puntos)
  for(i in 1:n_puntos){
    if(i==raiz){next}
    nivel_max <- niveles[i]
    coste_minimo <- 100000
    nodo_minimo <- NA
    #print(paste("El nivel maximo es ",nivel_max,sep=""))
    for(j in 1:n_puntos){
      if(j==i){next}
      #print(paste("El nivel del nodo ",j," es ",niveles[j]," y el nivel maximo es ",nivel_max,sep=""))
      if((niveles[j]<nivel_max)){
        #print("Primera restriccion pasada")
        if(costes[i,j]<coste_minimo){
          #print("Segunda restriccion pasada")
          coste_minimo <- costes[i,j]
          nodo_minimo <- j
        }
      }
    }
    resultado[i,nodo_minimo]=1
    resultado[nodo_minimo,i]=1
  }
  return(resultado)
}


get_Niveles <- function(arcos,raiz){
  # Primero en profundidad desde el nodo raíz manteniendo un contador de saltos
  n_nodos <- length(arcos[1,])
  nodos_visitados <- rep(0,n_nodos)
  solucion <- rep(0,n_puntos)
  # stack se tratará como una pila ( guarda los nodos a explorar)
  stack <- c(raiz)
  while(length(stack)!=0){
    nodo_actual <- stack[length(stack)]
    stack <- stack[-length(stack)]
    if(nodos_visitados[nodo_actual] == 0){
      nodos_visitados[nodo_actual] = 1
      for(i in which(arcos[nodo_actual,]==1) ){
        stack<-c(stack,i)
        if(nodos_visitados[i] == 0){
          solucion[i]<-solucion[nodo_actual]+1
        }
      }
    }
  }
  return(solucion)
}

shift1 <- function(solucion_actual,raiz){
  coste_actual <- sum(costes[which(solucion_actual==1)])/2
  niveles_actuales <- get_Niveles(solucion_actual,raiz)
  niveles_proximos <- niveles_actuales
  flag = 0
  i=raiz
  while(i==raiz){
    i <- sample(1:n_puntos, 1)
  }
  j = niveles_actuales[i]
  while(niveles_actuales[i] == j){
    j <- sample(1:H, 1)
    }
  niveles_actuales[i] = j
  solucion_prueba <- get_Grafo(niveles_actuales,costes)
  return(solucion_prueba)
}

swap <- function(solucion_actual){
  coste_actual <- sum(costes[which(solucion_actual==1)])/2
  niveles_actuales <- get_Niveles(solucion_actual,raiz)
  niveles_proximos <- niveles_actuales
  #Dibujar_grafo(cx,cy,solucion_actual)
  
  flag = 0
  while(flag == 0){
    flag = 1
    #print("Solucion actual")
    #print(niveles_actuales)
    for(i in 1:n_puntos){
      if(i==raiz){next}
      for(j in 1:n_puntos){
        if(j==raiz){next}
        if(niveles_actuales[i] == niveles_actuales[j]){next}
        niveles_prueba <- niveles_actuales
        var_paso <- niveles_prueba[i]
        niveles_prueba[i] = niveles_prueba[j]
        niveles_prueba[j] = var_paso
        solucion_prueba <- get_Grafo(niveles_prueba,costes)
        coste_prueba <- sum(costes[which(solucion_prueba==1)])/2
        if(coste_prueba < coste_actual){
          flag = 0
          coste_actual = coste_prueba
          niveles_proximos = niveles_prueba
          print(paste("Coste =",coste_actual,sep=""))
          #print(niveles_prueba)
        }
      }
    }
    niveles_actuales = niveles_proximos
    solucion_actual <- get_Grafo(niveles_actuales,costes)
    #Dibujar_grafo(cx,cy,solucion_actual)
  }
  return(solucion_actual)
}


sa <- function(TempActual,TempF,alpha,transiciones) {
  while(TempActual > TempF){
    t=0
    while(t<=transiciones){
      solucion_candidata <- shift1(solucion_actual,raiz)
      coste_candidato <- sum(costes[which(solucion_candidata==1)])/2
      mejora <- coste_actual - coste_candidato
      prob <- runif(1,0,1)
      if(mejora>0){
        solucion_actual <- solucion_candidata
        coste_actual <- coste_candidato
        #print(paste("Coste = ",coste_actual,sep=""))
        #print(paste("mejora = ",mejora,sep=""))
        #print(paste("Temperatura = ",TempActual,sep=""))
      }
      else if(exp(-abs(mejora)/TempActual) > prob){
        solucion_actual <- solucion_candidata
        coste_actual <- coste_candidato
        print(paste("Coste = ",coste_actual,sep=""))
        print(paste("Temperatura = ",TempActual,sep=""))
      }
      t=t+1
    }
    TempActual = TempActual * alpha
  }
  return(solucion_actual)
}
##############################################################################
##############################################################################


problemas<- read.csv("ejecuciones_2.txt",header = TRUE)
problemas <- problemas[1:4]


cat("archivo,problema,raiz,H,solución,tiemposa,tiempoew,centralidad_modificada,\n",file="sash/output.txt")

for(j in 1:length(problemas[,1])){
  print(problemas[j,])
  archivo = toString(problemas[j,1])
  problema = problemas[j,2] 
  raiz = problemas[j,3] 
  H = problemas[j,4] 
  # Datos por linea (Cada linea es un elemento del vector y son string)
  raw_data<-readLines(archivo,encoding="UTF-8")
  
  f1 <- function(x) unlist(strsplit(trim(x)," "))
  options(digits=7)
  # Datos por elemento (Cada número es un elemento del vector y son doubles)
  vector_data<-as.double(unlist(lapply(raw_data,f1)))
  
  n_problemas <- vector_data[1]
  vector_data <- vector_data[-1]
  indice=1
  cx<-c(); cy<-c()
  for(i in 1:n_problemas){
    n_puntos <- vector_data[indice]
    if(i == problema ){
      for(j in seq(1,2*n_puntos,2)){
        cx<-c(cx,vector_data[indice+j])
        cy<-c(cy,vector_data[indice+j+1])
      }
      break;
    }
    indice = indice + n_puntos*2 +1
  }
  costes <- as.matrix(dist(cbind(cx,cy)))
  
  # Coordenadas del problema escogido en:        cx cy
  # Costes en forma de distancia euclidea en:    costes
  
  # Centralidad midificada de la raiz
  centralidades <-c()
  for (i in 1:length(costes[,1])){
    centralidades <-c(centralidades,1/sum(costes[i,]))
  }
  nodomax <- which(centralidades == max(centralidades))
  nodomin <- which(centralidades == min(centralidades))
  centralidadm <- (centralidades[raiz] - centralidades[nodomin])/(centralidades[nodomax]-centralidades[nodomin])
  
  #############################################################################################
  ###########################      Simulated Annealing     ####################################
  #############################################################################################
  
  # arcs[i,j] =  1 si existe el arco si no = 0
  arcos <- matrix (0,n_puntos,n_puntos)
  # Solución inicial en forma de estrella
  arcos[raiz,(1:n_puntos)[-raiz]]=1
  arcos[(1:n_puntos)[-raiz],raiz]=1
  
  #Eliminación simple de los arcos (i,j) cuyo coste(i,j)>=coste(raiz,j)
  for(i in 1:n_puntos){
    if(i!=raiz){
      for(j in 1:n_puntos){
        if(j!=raiz){
          if(costes[i,j]>=costes[raiz,j]){
            arcos[i,j]=-1
            arcos[j,i]=-1
          }
        }
      }
    }
  }
  arcos_estrella <- arcos
  #Dibujar_grafo(cx,cy,arcos_estrella)
  
  ew_time <- proc.time()
  solucion_actual <- REW(arcos, 1)
  ew_time <- proc.time() - ew_time
  coste_actual <- sum(costes[which(solucion_actual==1)])/2
  #Dibujar_grafo(cx,cy,solucion_actual)
  
  
  TempActual = 1
  alpha = 0.999
  transiciones = 10
  TempF = 0.001
  sa_time <- proc.time()
  solucion_actual <- sa(TempActual,TempF,alpha,transiciones)
  sa_time  <- proc.time() - sa_time 
  
  coste_final <- sum(costes[which(solucion_actual==1)])/2
  
  arch <- strsplit(archivo, NULL)[[1]]
  arch<- paste(rev(arch),collapse = "")
  arch <- substr(arch,5,nchar(arch))
  arch <- strsplit(arch, NULL)[[1]]
  arch<- paste(rev(arch),collapse = "")
  
  print(paste("Coste sa ",coste_final,sep=""))
  #dib <- Dibujar_grafo(cx,cy,arcos)  
  fichero <- paste("sash/",arch,"_",problema,"_",raiz,"_",H,".png",sep="")
  png(filename=fichero)
  Dibujar_grafo(cx,cy,solucion_actual)
  dev.off()
  
  cat(paste(archivo,",",problema,",",raiz,",",H,",",round(coste_final, 3),",",round(sa_time[3],3),",",round(ew_time[3],3),",",round(centralidadm,3),"\n",sep=""),file="sash/output.txt",append = TRUE)
  
  }